;* --------------------------------------------------------------------*/
;*    Copyright (c) 1992-1998 by Manuel Serrano. All rights reserved.  */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \   /  '                               */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome. Send them to                                          */
;*        Manuel Serrano -- Manuel.Serrano@unice.fr                    */
;*-------------------------------------------------------------------- */
;*---------------------------------------------------------------------*/
;*    serrano/prgm/project/bigloo/runtime1.9/Rgc/dfa.scm               */
;*                                                                     */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Fri Apr  3 15:21:05 1992                          */
;*    Last change :  Wed Feb 19 13:27:07 1997 (serrano)                */
;*                                                                     */
;*    Le calcul des transitions du `dfa'                               */
;*---------------------------------------------------------------------*/
 
;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module __rgc_dfa

   (import  (__error                   "Llib/error.scm")
	    (__rgc                     "Rgc/runtime.scm")
	    (__rgc_tree                "Rgc/tree.scm"))
   
   (use     (__type                    "Llib/type.scm")
	    (__bigloo                  "Llib/bigloo.scm")
	    (__tvector                 "Llib/tvector.scm")
	    (__structure               "Llib/struct.scm")
	    (__tvector                 "Llib/tvector.scm")
	    (__r4_numbers_6_5          "Ieee/number.scm")
	    (__r4_numbers_6_5_fixnum   "Ieee/fixnum.scm")
	    (__r4_numbers_6_5_flonum   "Ieee/flonum.scm")
	    (__r4_characters_6_6       "Ieee/char.scm")
	    (__r4_equivalence_6_2      "Ieee/equiv.scm")
	    (__r4_booleans_6_1         "Ieee/boolean.scm")
	    (__r4_symbols_6_4          "Ieee/symbol.scm")
	    (__r4_strings_6_7          "Ieee/string.scm")
	    (__r4_pairs_and_lists_6_3  "Ieee/pair-list.scm")
	    (__r4_input_6_10_2         "Ieee/input.scm")
	    (__r4_control_features_6_9 "Ieee/control.scm")
	    (__r4_vectors_6_8          "Ieee/vector.scm")
	    (__r4_ports_6_10_1         "Ieee/port.scm")
	    (__r4_output_6_10_3        "Ieee/output.scm")
	    (__evenv                   "Eval/evenv.scm"))
  
   (export  (dfa tuple)))

;*---------------------------------------------------------------------*/
;*     dfa ...                                                         */
;*     ------------------------------------------------------------    */
;*     fast-union-v est passe en parametre car il a deja ete alloue    */
;*     (sa taille definitive est connue) par regular-grammar-2.        */
;*     ------------------------------------------------------------    */
;*     Toutes les unions triviales ne passent pas par les tables de    */
;*     hash mais sont retrouvees grace a un tableau (trivial).         */
;*     ------------------------------------------------------------    */
;*     t-alpha et l-alpha sont un tableau et une liste qui sont        */
;*     utilises pour calculer rapidement "lettre concernee a la pos".. */
;*     ------------------------------------------------------------    */
;*     l-trap est une liste qui contient toutes les traps. Une fois    */
;*     dstates calcule, on va gerer les traps. (passe trap)            */
;*---------------------------------------------------------------------*/
(define (dfa tuple)
   (let ((Dinit        (vector-ref tuple 0))
	 (position     (vector-ref tuple 1))
	 (f-store      (vector-ref tuple 2))
	 (f-env        (vector-ref tuple 3))
	 (fast-union-v (vector-ref tuple 4)))
      ;; on vide fast-union-v
      (fill-vector! fast-union-v #f)
      (let* ((Dstates-env   (make-env))
	     (Union-env     (make-env))
	     (nb-states-max 1024)
	     (nb-states     -1)
	     (states        (make-vector nb-states-max '()))
	     (P=a           '()) 
	     (t-alpha       (make-vector (+fx 1 *rgc-last-char*) '()))
	     (l-alpha       '())
	     (trivial       (make-vector (vector-length f-store) '())))
;*---- increment-nb-states! -------------------------------------------*/
	 (define (increment-nb-states!)
	    (set! nb-states (+fx 1 nb-states))
	    (if (=fx nb-states nb-states-max)
		(begin
		   (set! nb-states-max (* 2 nb-states-max))
		   (set! states (vector-grow states nb-states-max))))
	    nb-states)
;*---- make-state -----------------------------------------------------*/
	 (define (make-state symbol-name)
	    (let ((v (increment-nb-states!)))
	       (set! Dstates-env (cons (cons symbol-name v) Dstates-env))
	       nb-states))
;*---- set-alpha ------------------------------------------------------*/
;*   Si deux regles match une chaine, on ne prends que la 1ere action. */
;*   Pour modifier cela, il faut changer cette routine, ainsi que le   */
;*   code de main-loop a l'endroit on on fait:                         */
;*             (vector-set! dstates ... (cons a U) ...)                */
;*   (La modif decrite ci-dessus a ete effectuee.)                     */
;*---------------------------------------------------------------------*/
	 (define (set-alpha p*)
	    (set! l-alpha '())
	    (let loop ((p* (reverse p*)))
	       (if (null? p*)
		   '()
		   (let ((pr  (car p*))
			 (sp* (cdr p*)))
		      (let* ( (lettre (vector-ref position pr))
			      (alpha  (if (char? lettre)
					  lettre
					  'accept-position))
			      (indice (if (char? lettre)
					  (char->integer lettre)
					  0)))
			 (cond
			    ((null? (vector-ref t-alpha indice))
			     (set! l-alpha (cons alpha l-alpha))
			     (vector-set! t-alpha indice (cons pr '()))
			     (loop sp*))
			    (else
			     (vector-set! t-alpha 
					  indice 
					  (cons pr 
						(vector-ref t-alpha indice)))
			     (loop sp*))))))))
;*---- compute-real-union ---------------------------------------------*/
;*  Je garde cette fonction car je ne desepere pas de trouver une ruse */
;*  qui me permettrait une optimisation d'enfer...                     */
;*---------------------------------------------------------------------*/
	 (define (compute-real-union position*)
	    (labels ((first-non-null? (p* acc)
				      (if (null? p*)
					  (reverse! acc)
					  (if (null? (vector-ref f-store (car p*)))
					      (first-non-null? (cdr p*) acc)
					      (first-non-null? (cdr p*)
							       (cons (car p*)
								     acc))))))
	       (let ((p* (first-non-null? position* '())))
		  (if p*
		      (fast-union p* f-store fast-union-v)
		      '()))))
;*---------------------------------------------------------------------*/
;*     dfa                                                             */
;*---------------------------------------------------------------------*/
	 (let main-loop ((dstates (list
				   (cons Dinit
					 (make-state (make-state-name Dinit)))))) 
;*---- union-followpos ------------------------------------------------*/
;*  !!! WARNING !!!                                                    */
;*  ----------------------------------------------------------------   */
;*  C'est tres crade (mais efficace !), on fait un horrible            */
;*  side-effect sur dstates...                                         */
;*  ----------------------------------------------------------------   */
;*  On ne calcule pas union-followpos sur position* mais sur:          */
;*  (map f-env position*).                                             */
;*---------------------------------------------------------------------*/
	    (define (union-followpos position*)
	       (let ((env-pos (map (lambda (p) (vector-ref f-env p))
				   position*)))
;*---- La gestion des triviaux ----------------------------------------*/
		  (cond
		     ((null? (cdr env-pos))
		      (let ((indice (car env-pos)))
			 (if (null? (vector-ref f-store indice))
			     '()
			     (begin
				(if (null? (vector-ref trivial indice))
				    (let ((state-name
					   (make-state-name
					    (vector-ref
					     f-store indice))) )
				       (let ((num (bound?
						   state-name
						   Dstates-env)))
					  (if num
					      (begin
						 (vector-set! trivial
							      indice
							      num)
						 num)
					      (let ((num (make-state
							  state-name)))
						 (vector-set! trivial
							      indice
							      num)
						 (set! dstates
						       (cons
							(cons (vector-ref
							       f-store
							       indice)
							      num) 
							dstates))
						 num))))
				    (vector-ref trivial indice))))))
;*---- Les cas non-triviaux -------------------------------------------*/
		     (else
		      (let ((union-name (make-union-name env-pos)))
			 (let ((num (bound? union-name Union-env)))
			    (if num
				num
				(let* ((U (compute-real-union env-pos))
				       (state-name (make-state-name U)))
				   (let ((num (bound? state-name Dstates-env)))
				      (if num
					  (begin
					     (set! Union-env
						   (cons
						    (cons union-name num)
						    Union-env))
					     num)
					  (let ((num (make-state state-name)))
					     (set! dstates
						   (cons (cons U num)
							 dstates))
					     (set! Union-env
						   (cons
						    (cons union-name num)
						    Union-env))
					     num)))))))))))
;*---- main-loop ------------------------------------------------------*/
	       (if (null? dstates)
		   ;; on a fini le calcul des transitions
		   (cons (+fx 1 nb-states) states)
		   ;; Il reste des etats non marques
		   (let* ((T    (car (car dstates)))
			  (Tnum (cdr (car dstates))))
		      ;; on met en place t-alpha et l-alpha
		      (set-alpha T)
		      ;; Ceci revients a marquer dstates
		      (set! dstates (cdr dstates))
		      (let loop ((a* l-alpha))
			 (if (null? a*)
			     (main-loop dstates)
			     (let ((a (car a*)))
				(cond
				   ((char? a)
				    ;; Ce n'est pas une lettre acceptante
				    (let ( (indice (char->integer a)) )
				       (set! P=a
					     (vector-ref t-alpha indice))
				       (vector-set! t-alpha indice '())
				       (let ((U (union-followpos P=a)))
					  (vector-set! 
					   states
					   Tnum
					   (cons
					    (cons a U)
					    (vector-ref states Tnum))))))
				   (else
				    ;; C'est une lettre acceptante
				    (let ((indice 0)
					  (accept
					   (remq!
					    'epsilon
					    (map (lambda (p)
						    (vector-ref position p))
						 (vector-ref t-alpha 0)))))
				       (vector-set! t-alpha 0 '())
				       (if (not (null? accept))
					   (vector-set!
					    states
					    Tnum
					    (cons accept
						  (vector-ref states
							      Tnum)))))))
				(loop (cdr a*))))))))))))

;*---------------------------------------------------------------------*/
;*     make-prefix-name ...                                            */
;*---------------------------------------------------------------------*/
(define (make-prefix-name prefix num*)
   (let ((s (apply
	     string-append
	     (cons prefix
		   (map (lambda (num)
			   (string-append "." (integer->string num)))
			num*)))))
      (string->symbol
       (string-upcase
	(apply
	 string-append
	 (cons prefix
	       (map (lambda (num)
		       (string-append "." (integer->string num)))
		    num*)))))))

;*---------------------------------------------------------------------*/
;*     bound? ...                                                      */
;*---------------------------------------------------------------------*/
(define (bound? name env)
   (let ((bb (assq name env)))
      (if bb (cdr bb) #f)))
		  
;*---------------------------------------------------------------------*/
;*     set-in-env! ...                                                 */
;*---------------------------------------------------------------------*/
(define-inline (set-in-env! name val env)
   (let ((b (assq name env))
	 (v val) )
      (set-cdr! b v)
      v))
  
;*---------------------------------------------------------------------*/
;*     make-env ...                                                    */
;*---------------------------------------------------------------------*/
(define-inline (make-env)
   '())

;*---------------------------------------------------------------------*/
;*     make-state-name ...                                             */
;*---------------------------------------------------------------------*/
(define-inline (make-state-name num*)
   (make-prefix-name "state" num*))

;*---------------------------------------------------------------------*/
;*     make-union-name ...                                             */
;*---------------------------------------------------------------------*/
(define-inline (make-union-name num*)
   (make-prefix-name "union" num*))

;*---------------------------------------------------------------------*/
;*     fast-union                                                      */
;*     ------------------------------------------------------------    */
;*     L'indirection f-env a deja ete faite dans union-followpos. il   */
;*     ne reste donc a faire que celle sur f-store.                    */
;*---------------------------------------------------------------------*/
(define (fast-union l* f-store fast-union-v)
   (if (null? (cdr l*))
       (vector-ref f-store (car l*))
       (let* ((init (car (vector-ref f-store (car l*))))
	      (max  init)
	      (min  init))
;*---- On lit toutes les listes ---------------------------------------*/
	  (labels ((read (l)
			 (if (null? l)
			     '()
			     (let ( (c (car l)) )
				(cond
				   ((<fx c min)
				    (set! min c))
				   ((>fx c max)
				    (set! max c)))
				(vector-set! fast-union-v c #t)
				(read (cdr l))))))
	     (let loop ((l l*))
		(if (null? l)
		    'read-done
		    (begin
		       (read (vector-ref f-store (car l)))
		       (loop (cdr l))))))
;*---- on ecrit le resultat -------------------------------------------*/
	  (let loop ((i   max)
		     (acc '()))
	     (cond
		((<fx i min)
		 acc)
		((vector-ref fast-union-v i)
		 (vector-set! fast-union-v i #f)
		 (loop (-fx i 1) (cons i acc)))
		(else
		 (loop (-fx i 1) acc)))))))

;*---------------------------------------------------------------------*/
;*    fill-vector! ...                                                 */
;*---------------------------------------------------------------------*/
(define (fill-vector! vector fill)
   (let loop ((i (-fx (vector-length vector) 1)))
      (if (=fx i -1)
	  'done
	  (begin
	     (vector-set! vector i fill)
	     (loop (-fx i 1))))))

